# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
"""
Contains the top-level Window and global execution environment.
"""
import random
import asyncio
import time
import pygame #todo try
import pygame.key
from pygame.locals import *

import scratchypy.stage
from scratchypy import color, util

class _RollingAverage:
    """
    Simple class to keep a circular buffer of the last N values and
    calculate a rolling average on them.
    The buffer starts out as zeros so that will skew the results at the 
    beginning.
    """
    def __init__(self, numValues=10):
        if numValues <= 0:
            raise ValueError("must be positive")
        self._values = [0 for _ in range(numValues)]
        self._numValues = numValues
        self._index = 0
        
    def append(self, value):
        self._values[self._index] = value
        self._index = (self._index + 1) % self._numValues
        
    def average(self):
        return sum(self._values) / self._numValues
    

class Window:
    FPS=30
    FRAME_SEC = 1 / FPS
    
    def __init__(self):
        self._stage = scratchypy.stage.Stage()
        self._mousePos = (0,0)
        self._mouseDown = False
        self._windowSize = (800,600)
        self._fullScreen = False
        self._backgroundColor = color.WHITE
        self._running = False
        self._rollingFrameSec = _RollingAverage()
        self._lastDraw = time.perf_counter() # high resolution timer
        self._debug = False
        self._epoch = time.monotonic()
        
    def set_debug(self, val=True):
        self._debug = val
        
    def set_size(self, width, height):
        """
        """
        if self._running:
            raise RuntimeError("Can only be set before running")
        self._windowSize=(width, height)
        
    def set_fullscreen(self):
        self._fullScreen = True
        
    def set_background_color(self, color:pygame.color.Color):
        self._backgroundColor = color

    @property
    def stage(self):
        return self._stage
    
    def set_stage(self, newStage):
        #TODO: more to it than this?
        # cleanup old, but some events might still reference
        self._stage = newStage
        self._stage._start()
    
    @property
    def fps(self) -> int:
        """
        The target Frames Per Second used for calculations.
        """
        return self.FPS
    
    @property
    def actual_fps(self) -> float:
        """
        @return the actual calculated FPS as a float.  This is usually slightly
        less than the configured fps() due to overhead.  But if you see it drop,
        you may be having a performance problem or taking too much time in the
        event callbacks.
        """
        return 1 / self._rollingFrameSec.average()

    @property
    def mouse_x(self):
        return self._mousePos[0]

    @property
    def mouse_y(self):
        return self._mousePos[1]

    @property
    def mouse_pointer(self):
        "@return the position of the mouse pointer as an (x,y) tuple."
        return self._mousePos

    #not a property?
    def mouse_down(self):
        return self._mouseDown

    def key_pressed(self, key=None):
        """
        Return True if the given key is currently pressed.  If no key given, then
        returns True if any key is pressed.
        @param key: Can be the text name of the key.  Some examples are 'a', '3',
               'space', 'up', 'down', 'left', 'right', etc.  The complete list of 
               names is determined by the pygame.key module.
               Key can also be a pygame.K_* integer constant.
        """
        boollist = pygame.key.get_pressed()
        if key is None or key == 'any':
            return any(boollist)
        elif isinstance(key, str):
            intkey = pygame.key.key_code(key) #throw if not found
            return boollist[intkey]
        elif isinstance(key, int) and key >= 0 and key < len(boollist):
            # pygame K_* constant
            return boollist[key]
        else:
            raise TypeError("key_pressed key not recognized: '%s'" % str(key))

    @property
    def random_position(self):
        """
        Pick a random position within the window and return it as an (x,y) pair.
        @return a tuple of an (x,y) coordinate
        """
        x = random.randint(0, self._windowSize[0])
        y = random.randint(0, self._windowSize[1])
        return (x,y)

    @property
    def size(self):
        "@return the (w,h) size of the window as a tuple"
        return self._windowSize

    @property
    def rect(self):
        return pygame.Rect(0,0, *self._windowSize)
    
    @property
    def timer(self):
        """
        Like the Scratch 'timer' pseudo-variable, this will return the number
        of seconds (and fractional seconds) since the program was started.
        """
        return time.monotonic() - self._epoch
    
    def reset_timer(self):
        self._epoch = time.monotonic()

    def _make_screen(self, windowSize):
        winstyle = pygame.FULLSCREEN if self._fullScreen else 0
        bestdepth = pygame.display.mode_ok(self._windowSize, winstyle, 32)
        screen = pygame.display.set_mode(self._windowSize, winstyle, bestdepth)
        return screen

    def _handleEvents(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                raise StopIteration() #TODO
            elif event.type == pygame.MOUSEBUTTONDOWN:
                self._mouseDown = True
                self._stage._on_mouse_down(event)
            elif event.type == pygame.MOUSEBUTTONUP:
                self._mouseDown = False
                self._stage._on_mouse_up(event)
            elif event.type == pygame.MOUSEMOTION:
                self._mousePos = event.pos
                self._stage._on_mouse_motion(event)
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    raise StopIteration() #TODO
                else:
                    self._stage._on_key_down(event)

    def _async_tick(self, screen):
        # paint the screen
        pygame.display.flip()
        # stamp when the screen was last painted and add to rolling average
        now = time.perf_counter()  # high resolution timer
        elapsed = now - self._lastDraw
        self._rollingFrameSec.append(elapsed)
        self._lastDraw = now
        fudge = max(0, elapsed - self.FRAME_SEC) # overhead time
        # Reschedule a draw for later.
        # This limits the framerate, like tick(FPS), but also gives async
        # callbacks triggered by below events a chance to run within the same
        # frame.
        againHandle = asyncio.get_running_loop().call_later(self.FRAME_SEC - fudge, self._async_tick, screen)
        
        try:
            self._handleEvents()
            screen.fill(self._backgroundColor)
            self._stage._update(screen)
        except StopIteration:
            againHandle.cancel()
            asyncio.get_running_loop().stop()
            return
        except Exception as ex:
            print("Exception from events: '%s'." % str(ex))
            #TODO: stop and show dialog?
        
    def run(self):
        screen = self._make_screen(self._windowSize)
        util.set_ui_thread()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_debug(self._debug)
        loop.slow_callback_duration = 1 / self.fps
        loop.call_soon(self._stage._start)
        loop.call_soon(self._async_tick, screen)
        loop.run_forever()
        # drain cancellations
        #print("Cancelling tasks")
        #for task in asyncio.all_tasks():
            #task.cancel()
        try:
            loop.run_until_complete(asyncio.gather(*asyncio.Task.all_tasks()))
        except:
            pass #TODO: other try/except may spew on exit
        loop.close()
        
    async def next_frame(self):
        """ 
        Yields control until the next frame is drawn.  This is done by a 
        calculated sleep so isn't 100% guaranteed. TODO
        """
        await asyncio.sleep(self.FRAME_SEC)

## Module functions
_window = Window()
def get_window():
    """
    @return the one and only window.  None if start() hasn't been called.
    """
    return _window

def get_stage():
    return get_window().stage

def set_stage(newStage):
    """
    Convenience to set the current stage on the window.
    """
    get_window().set_stage(newStage)

def start(whenStarted=None, stage=None, windowSize=None, fullScreen=False, backgroundColor=None, asyncioDebug=False):
    """
    Shows the window and starts the event loop.  Never returns.
    There are many options that are all optional.  It is best to
    always set these using the keyword parameters.
    @param whenStarted A function to call to perform initialization.
           This is usually only used if you are doing a simple
           functional program.  If you're organizing using custom
           stages, then you'd probably make your sprites in 
           stage.on_init().
    @param stage If you make a custom stage, you can supply it here
           and it will be set as the window's stage.
    @param windowSize A tuple of (width, height) for how big to make the window.
           Default (800, 600).
    @param fullScreen If true, use the full screen and windowSize is ignored.
    @param backgroundColor A scratchypy.color or pygame.color to use as the 
           window background, default WHITE.
    @param asyncioDebug Advanced logging of Python asyncio calls.
    """
    global _window
    if windowSize:
        _window.set_size(*windowSize)
    if fullScreen:
        _window.set_fullscreen()
    if backgroundColor:
        _window.set_background_color(backgroundColor)
    if stage is not None:
        _window._stage = stage # don't start it yet until loop is made
    if whenStarted:
        _window.stage.when_started(whenStarted)
    if asyncioDebug:
        _window.set_debug(True)
    _window.run()  #forever



# Copyright 2023 Mark Malek
# See LICENSE file for full license terms.
"""
This is the documentation for the ScratchyPy library.  I suggest looking at
the `sprite` and `stage` modules first! 
""" 

import pygame

pygame.init()

from .version import __version__
print("ScratchyPy " + __version__)

from . import sprite, stage, window, color, sound, image, text
from .window import *
from .stage import *
from .sprite import *
from .util import *



# Copyright 2022 Mark Malek
# See LICENSE file for full license terms.
# You are free to use this example in your own code.


# This line tells Python that we can use everything (*) in ScratchyPy directly
# in this file.
from scratchypy import *

# This is how to define a function in Python.  The name can be anything you'd
# like (subject to the Python naming rules, of course).
# This function will receive a `stage` that we can use to draw on.
def startItUp(stage):
    # This creates a new Sprite called 'sprite1'.  The costume for the sprite
    # is loaded from the given image file.  Use your favorite paint program to
    # make your image, and see the documentation for tips on how to save.
    # The sprite will be centered in the window by default.
    sprite1 = Sprite("assets/axolotl1.png")
    # Sprites are not automatically added to the stage.  Do so like this.
    stage.add(sprite1)
    # Rotate the sprite 45 degrees clockwise.
    sprite1.turn(45)
    # Make the sprite travel 50 steps (pixels) in the direction it is facing.
    # Like Scratch, the sprite by default faces to the right so this will move 
    # it diagonally down-right 50 pixels.
    sprite1.move(50)
    
# This is like the "when flag clicked" block, except there is no flag button
# in ScratchyPy to start the program - it starts automatically.  This tells
# the program to run the 'startItUp() function above when the program starts.
start(startItUp)


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 

import sys
sys.path.append("..")
from scratchypy import *
from pygame.locals import *
import pygame.font

#globals
rotatingText = None
changingTextLeft = None
changingTextRight = None
counter = 0

async def init(stage):
    global rotatingText, changingTextLeft, changingTextRight
    
    stage.forever(tick)

    wrapText = TextSprite("This is some text that is really long and wraps",
                        size=40, topleft=(0, 10), maxWidth=410)
    stage.add(wrapText)
    
    rotatingText = TextSprite("Rotating text", size=40, topleft=(0, 90))
    stage.add(rotatingText)
    
    changingTextLeft = TextSprite("Text changing left:", size=40, maxWidth=500, topleft=(0, 180))
    stage.add(changingTextLeft)
    changingTextRight = TextSprite("Text changing right:", size=40, maxWidth=500, topright=(400, 220))
    stage.add(changingTextRight)
    
    blueText = TextSprite("Blue text", size=40, color=color.BLUE, topleft=(0,260))
    stage.add(blueText)
    
    smallText = TextSprite("Small text", size=12, color=color.RED, topleft=(0,300))
    stage.add(smallText)
    
    font = pygame.font.SysFont("serif", 40, italic=True)
    fontText = TextSprite("Different font", font=font, topleft=(0, 320))
    stage.add(fontText)
    
    highlightText = TextSprite("highlight", size=40, bgcolor=color.YELLOW, topleft=(0,360))
    stage.add(highlightText)
    
    highlightText2 = TextSprite("highlight with wrap " * 6, size=30, bgcolor=color.GREEN, topleft=(0,400))
    stage.add(highlightText2)
    
    multiLine = TextSprite("multiline1 with some wrapping stuff center justified\nline2\nline3", size=30, maxWidth=320, topleft=(0,460), justification='center')
    stage.add(multiLine)

def tick(stage):
    #print("tick")
    global rotatingText, changingTextLeft, changingTextRight, counter
    rotatingText.turn(2)
    if counter % 30 == 0:
        changingTextLeft.set_text("Text changing left: %d" % (counter//30))
        changingTextRight.set_text("Text changing right: %d" % (counter//30))
    counter += 1


if __name__ == '__main__':
    start(whenStarted=init)
    
    
    


# Copyright 2023 Mark Malek
# See LICENSE file for full license terms. 
"""
Contains the Sprite class, the base of all sprites on a stage, as well as 
several other specialized subclasses.
"""

import pygame.sprite
import pygame.font
import math
import asyncio
import inspect
import copy
from typing import Literal, Tuple, Union
from scratchypy.window import get_window
from scratchypy import color
from scratchypy.eventcallback import EventCallback
import scratchypy.text


# Rotation styles
DONT_ROTATE = 0
LEFT_RIGHT = 1
ALL_AROUND = 2

# For making unique sprite names
_idCounter = 0

class Sprite(pygame.sprite.Sprite): 
    
    def __init__(self, costumes,
                 x=None, y=None, topleft=None, topright=None,
                 name=None, size=None, stage=None):
        """
        Like Scratch, x,y is in the center of the sprite.
        Where that is may change based on the costume.
        """
        #TODO: separate from pygame groups
        self.groups = [ stage ] if stage else []
        pygame.sprite.Sprite.__init__(self, self.groups)
        
        global _idCounter
        _idCounter += 1
        self._name = name if name else "sprite" + str(_idCounter)
        
        self._costumes = []  #TODO ordered dictionary so we can lookup by name too
        self._costumeIndex = 0 #TODO does not exist yet
        self._image = None # set in _applyImage
        self._mask = None  #ditto
        self._rect = None # ditto
        self._visible = True
        winx, winy = get_window().size
        self._x = x if x is not None else winx/2
        self._y = y if y is not None else winy/2
        
        self._scale = 1 # 0..1..n, but API uses 0..100%..n%
        if size is not None and size >= 0: #TODO confusing name, but matches scratch
            self._scale = size/100
        self._rotation = 0  # degrees clockwise, like Scratch
        self._rotationStyle = ALL_AROUND # matches Scratch modes, may flip
        
        #TODO
        self._draggable = False
        self._sayThinkImages = None # images (right,left) when saying or thinking
        self._debug = False
    
        # Events
        self._on_click = EventCallback(self, None)
        self._messageHandlers = {}
        self._keyHandlers = {} # Dict of key->EventCallback
        self._on_tick = EventCallback(self, None)
        
        # TODO: have a mode to keep on screen
        self._loadCostumes(costumes if isinstance(costumes, list) else [ costumes ])
        #TODO: assert at least one costume
        
        #Easy way to position by topleft/topright instea
        if topleft:
            self.go_rect(topleft=topleft)
        elif topright:
            self.go_rect(topright=topright)
            
        if stage is not None:
            stage.add(self)
            
        # Important: if you add any new members, you must take a look at clone()
        
    def _loadCostumes(self, listOfImages):
        for im in listOfImages:
            if isinstance(im, str):
                image = pygame.image.load(im)
                image.convert(32, pygame.SRCALPHA)
            elif isinstance(im, pygame.Surface):
                image = im
            self._costumes.append(image)
        self.switch_costume_to(0)
        
    def _applyImage(self):
        """ Apply scales and transforms to original image, then set sprite vars """
        # These variables needed for sprite conventions
        self._image = orig = self._costumes[self._costumeIndex]
        if self._rotationStyle == LEFT_RIGHT and self._rotation >= 180:
            self._image = pygame.transform.flip(self._image, True, False)
        elif self._rotationStyle == ALL_AROUND and self._rotation:
            #self._image = pygame.transform.rotozoom(self._image, -self._rotation, self._scale) # pygame is CCW
            self._image = pygame.transform.rotate(self._image, -self._rotation) # pygame is CCW
        # else no rotation or flip
        
        if self._scale:
            r = self._image.get_rect()
            newSize = (int(r.width * self._scale), int(r.height * self._scale))
            self._image = pygame.transform.smoothscale(self._image, newSize)

        self._image.set_colorkey(orig.get_colorkey()) 
        self._mask = pygame.mask.from_surface(self._image)
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def _on_mouse_motion(self, event):
        if self._draggable:
            pass #TODO
    
    @property
    def name(self):
        """
        @return the name of the sprite set in the constructor
        """
        return self._name
    
    @property
    def visible(self):
        """
        @return True if visible, or False if hide() has been called.
        """
        return self._visible
    
    @property
    def rect(self):
        return self._rect
    
    #################################################
    ##                  MOTION
    #################################################
    
    def move(self, steps:float):
        """
        Move C{steps} (pixel lengths) in the direction of the current
        rotation.
        @param steps Distance to travel.  May be a fraction.
        FIXME for left-right mode
        """
        dy = steps * math.sin(self._rotation * math.pi / 180)
        dx = steps * math.cos(self._rotation * math.pi / 180)
        self._y += dy
        self._x += dx
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def turn(self, degrees:float):
        """
        Change the direction by the amouny of degrees given in C{degrees}.
        A positive number is clockwise 🔃.
        A negative number is counter-clockwise 🔄.
        """
        self._rotation += degrees
        self._rotation %= 360
        self._applyImage()
        
    def go_to_position(self, position:Tuple[float,float]):
        """
        Go to the (x,y) postion, given as a tuple.
        @see scratchypy.random_position()
        @see scratchypy.mouse_pointer()
        """
        self.goTo(position[0], position[1])
        
    def go_to(self, x:float, y:float):
        self._x = x
        self._y = y
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def go_rect(self, **kwargs):
        """
        Go to a position determined by the keywords accepted by pygame.Rect.
        """
        self._rect = self._image.get_rect(**kwargs)
        self._x = self._rect.centerx
        self._y = self._rect.centery
        
    async def glide_to_and_wait(self, x:float, y:float, seconds:float):
        nframes = get_window().fps * seconds
        while nframes > 0:
            # Recalculate each time in case another event moved us
            dx = (x - self._x) / nframes
            dy = (y - self._y) / nframes
            self.change_x_by(dx)
            self.change_y_by(dy)
            nframes -= 1
            await get_window().next_frame()
        # When done, should be at final spot
        self.go_to(x, y)
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
        
    def glide_to(self, x:float, y:float, seconds:float):
        """
        Glide to the given x, y in the background without waiting.
        @return The task object which can be waited on for completion.
        """
        return asyncio.create_task(self.glide_to_and_wait(x, y, seconds))
    
    async def glide_to_position_and_wait(self, position:Tuple[float,float], seconds:float):
        "Similar to glide_to_and_wait but with single position argument"
        await self.glide_to_and_wait(position[0], position[1], seconds)
    
    def glide_to_position(self, position:Tuple[float,float], seconds:float):
        "Similar to glide_to but with single position argument"
        return self.glide_to(position[0], position[1], seconds)
        
    def point_in_direction(self, degrees:float):
        """
        Set the *direction* to the given C{degrees}.  The sprite will
        point that direction regardless of its current direction.
        """
        self._rotation = (degrees - 90) % 360
        self._applyImage()
        
    def point_towards(self, positionXY):
        if isinstance(positionXY, Sprite):
            posX = positionXY.x
            posY = positionXY.y
        else:
            posX = positionXY[0]
            posY = positionXY[1]

        # The rotation(direction) is calculated no matter the mode.
        # Instead, _applyImage() decides whether to draw the rotation.
        dy = posY - self._y # may be negatives
        dx = posX - self._x
        if dx == 0:  # straight up or down
            self._rotation = 0 if dy >= 0 else 180
        else:
            self._rotation = math.atan(dy/dx) * 180 / math.pi
            if dx < 0:
                self._rotation += 180
        self._applyImage()
            
    def change_x_by(self, steps):
        self._x += steps
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def set_x_to(self, x):
        self._x = x  # TODO: snap to screen?
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def change_y_by(self, steps):
        self._y += steps
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def set_y_to(self, y):
        self._y = y
        self._rect = self._image.get_rect(center=(self._x, self._y))
        
    def if_on_edge_bounce(self):
        """
        TODO: add padding
        """
        winx,winy = get_window().size
        if self._rect.left < 0:
            self._x = self._rect.w / 2
            self.point_in_direction(-self.direction)
        if self._rect.right >= winx:
            self._x = winx - self.rect.w / 2
            self.point_in_direction(-self.direction)
        if self._rect.top < 0:
            self._y = self._rect.h / 2
            self.point_in_direction(180-self.direction)
        if self._rect.bottom >= winy:
            self._y = winy - self._rect.h / 2
            self.point_in_direction(180-self.direction)
            
    def if_on_edge_snap(self, padding=0):
        pass #TODO
            
    def set_rotation_style(self, style:Literal[DONT_ROTATE, LEFT_RIGHT, ALL_AROUND]):
        """ Use one of the enumerated values. """
        if not style in (DONT_ROTATE, LEFT_RIGHT, ALL_AROUND):
            raise ValueError("bad style type")
        self._rotationStyle = style
        if style in (DONT_ROTATE, ALL_AROUND):
            self._rotation = 0
        self._applyImage()
        
    @property
    def x_position(self) -> float:
        """
        sprite.x is an alias for this.
        @return the x coordinate of the center of this sprite.
        """
        return self._x
    x = x_position
    
    @property
    def y_position(self) -> float:
        """
        sprite.y is an alias for this.
        @return the y coordinate of the center of this sprite.
        """
        return self._y
    y = y_position
    
    @property
    def position(self) -> Tuple[float,float]:
        return (self._x, self._y)
    
    @property
    def direction(self) -> float:
        """
        The "direction" in Scratch is the angle the character is facing.
        By default, the cat faces right, 90 degrees - that is equivalent
        to not being rotated (0 degrees).
        Direction is in the range -179.99.. to +180 inclusive.
        """
        d = (self._rotation + 270) % 360 - 180  
        return d if d != -180 else 180   # prefer positive 180
            
    #################################################
    ##                  LOOKS
    #################################################

    def _make_bubble(self, speechText:str):
        #render the text
        font = pygame.font.SysFont("sans serif", 25)
        bounds = pygame.Rect(0,0,320,320)
        textSurf = scratchypy.text.render_text(font, speechText, bounds)
        
        # Draw the bubble - TODO-handle min sizes
        textRect = textSurf.get_rect().move(4,4) #adjust so can inflate
        bubbleRect = textRect.inflate(8,8)
        
        bubbleSurf = pygame.surface.Surface((bubbleRect.w, bubbleRect.h+10), pygame.SRCALPHA, 32)
        pygame.draw.rect(bubbleSurf, color.WHITE, bubbleRect, width=0, border_radius=10)
        pygame.draw.rect(bubbleSurf, color.BLACK, bubbleRect, width=1, border_radius=10)
        bubbleSurf.blit(textSurf, textRect)
        return bubbleSurf

    def say(self, speechText:str=None):
        """ 
        Will stay indefinitely or until another say/think.
        Say None or blank to clear bubble.
        """
        if not speechText:
            self._sayThinkImages = None
            return
        
        right = self._make_bubble(speechText)
        left = right.copy()
        r = right.get_rect()
        # draw tail for right orientation (left side)
        points = [ (15, r.bottom-11),
                   (10, r.bottom),
                   (22, r.bottom-11) ]
        pygame.draw.polygon(right, color.WHITE, points, width=0)
        pygame.draw.lines(right, color.BLACK, False, points, width=1)
        # draw tail for left orientation (right side)
        points = [(r.right-x, y) for x,y in points]
        pygame.draw.polygon(left, color.WHITE, points, width=0)
        pygame.draw.lines(left, color.BLACK, False, points, width=1)
        # set it
        self._sayThinkImages = (right, left)
        
    async def say_and_wait(self, speechText:str, howManySeconds:float):
        self.say(speechText)
        await asyncio.sleep(howManySeconds)
        self._sayThinkImages = None
        
    def think(self, thoughtText:str):
        """ 
        Will stay indefinitely or until another say/think.
        Say None or blank to clear bubble.
        """
        if not thoughtText:
            self._sayThinkImages = None
            return
        
        right = self._make_bubble(thoughtText)
        left = right.copy()
        r = right.get_rect()
        # right side (left bubbles) 
        pygame.draw.circle(right, color.WHITE, (16, r.bottom-9), 4, width=0)
        pygame.draw.circle(right, color.BLACK, (16, r.bottom-9), 4, width=1)
        pygame.draw.circle(right, color.WHITE, (11, r.bottom-4), 2, width=0)
        pygame.draw.circle(right, color.BLACK, (11, r.bottom-4), 2, width=1)
        # left side (right bubbles) 
        pygame.draw.circle(left, color.WHITE, (r.right-16, r.bottom-9), 4, width=0)
        pygame.draw.circle(left, color.BLACK, (r.right-16, r.bottom-9), 4, width=1)
        pygame.draw.circle(left, color.WHITE, (r.right-11, r.bottom-4), 2, width=0)
        pygame.draw.circle(left, color.BLACK, (r.right-11, r.bottom-4), 2, width=1)
        # set it
        self._sayThinkImages = (right, left)
        
    async def think_and_wait(self, thoughtText:str, howManySeconds:float):
        self.think(thoughtText)
        await asyncio.sleep(howManySeconds)
        self._sayThinkImages = None
    
    def switch_costume_to(self, index:int):
        """
        TODO: support switch by name (file name?)
        @param number The index (number) of the costume to select.
                      Remember, in Python, counting starts at 0, 1, 2, ...
        """
        if not self._costumes:
            return #BOOM
        if index < 0 or index > len(self._costumes):
            index = 0
        self._costumeIndex = index
        self._applyImage()
        
    def next_costume(self):
        self._costumeIndex = (self._costumeIndex + 1) % len(self._costumes)
        self._applyImage()
        
    # switch_backdrop_to() is on the stage object
    # next_backdrop() is on the stage object
    
    def change_size_by(self, percentChange:float):
        #TODO: don't let less than 5% or n pixels or x% bigger than screen
        self._scale = self._scale + percentChange/100
        if self._scale < 0.05:
            self._scale = 0.05
        self._applyImage()
    
    def set_size_to(self, percent:float):
        #TODO: don't let less than 5% or n pixels or x% bigger than screen
        if percent < 0:
            raise ValueError("setSizeTo(scale) must be >= 0")
        self._scale = percent/100
        self._applyImage()
    
    #TODO: color effects
        
    def show(self):
        self._visible = True
        
    def hide(self):
        self._visible = False
        
    #TODO layers
    
    @property
    def costume_number(self) -> int:
        return self._costumeIndex
    
    @property
    def costume_name(self) -> str:
        return "TODO: implement me"
    
    # backdrop on stage
    
    @property
    def size(self) -> float:
        """
        @return the size (scaling percent) as a number, where 100% is full size.
        """
        return self._scale * 100
    
    #################################################
    ##                  SOUND - should be a separate module
    #################################################
    
    #################################################
    ##                  EVENTS
    #################################################
    
    # when flag clicked = whenStarted on start()
    
    def when_key_pressed(self, key, functionToCall):
        if not inspect.isfunction(functionToCall):
            raise TypeError("callback is not a function")
        if isinstance(key, str):
            intkey = pygame.key.key_code(key) #throw if not found
        elif isinstance(key, int):
            intkey = key
        else:
            raise TypeError("unknown key")
        name = pygame.key.name(intkey) + " key handler: " + functionToCall.__name__
        self._keyHandlers[intkey] = EventCallback(self, functionToCall, name=name)
        
    def _on_key_down(self, event):
        handler = self._keyHandlers.get(event.key)
        if handler:
            handler()
    
    def when_clicked(self, handler):
        """
        When this sprite is clicked, call the given handler.
        The handler is of the form:
        ```
        def myClickHandler(sprite, clickPosition):
            pass
        ```
        Where 'sprite' is this sprite object where the click was delivered, and
        clickPosition is the tuple (x,y) position of the click relative to the
        sprite.
        """
        self._on_click.set(handler)
    when_this_sprite_clicked = when_clicked
    
    def when_i_receive(self, messageName:str, handlerFunction):
        """
        Register a callback to call when this sprite receives a message with
        the given messageName.
        The handler is of the form:
        ```
        def myMessageHandler(sprite, argDictionary):
            pass
        ```
        Where 'sprite' is this sprite object where the click was delivered, and
        the argDictionary is whatever kind of extra information goes with the
        message, as a dictionary of key, value pairs.
        """
        self._messageHandlers[messageName] = EventCallback(self, handlerFunction)
    
    def message(self, messageName:str, argDictionary={}):
        """
        Sends a message to this sprite.  (Bonus extension of Scratch).
        If there is no message event handler (registered with when_i_receive()),
        for the given messageName, then the message is ignored.
        @param messageName The message name, as a string
        @param argDictionary Optional extra parameters to give as part of the message
        FIXME: this should post to a queue, not run handler directly?
        """
        handler = self._messageHandlers.get(messageName)
        if handler:
            handler(argDictionary)
        elif self._debug:
            print("%s: no handler found for message %s" % (self._name, messageName))
    
    def broadcast(self, messageName, **kwargs):
        pass  #TODO - in stage
    
    # broadcast_and_wait????  Supposed to wait until all handlers complete
    
    #################################################
    ##                  CONTROL
    #################################################
    def forever(self, functionToCall):
        self._on_tick.set(functionToCall)
        
    def clone(self, name=None, stage=None):
        """
        Make a shallow clone of this sprite.
        Important: The cloned sprite is not automatically on the same stage as
        the original; specify the 'stage' parameter to add the sprite to the
        stage.
        Costumes are not reloaded but can be manipulated independently.  
        The new sprite is at the same location and with the all the same 
        properties as this sprite. You can change properties or call functions 
        on the new returned object.
        @param name A unique name for the new sprite, or one will be
               automatically assigned.
        @param stage If not None, will be automatically added to the stage.
        """
        newObj = copy.copy(self)
        global _idCounter
        _idCounter += 1
        newObj._name = name if name else "sprite" + str(_idCounter)
        newObj._costumes = self._costumes.copy()
        newObj.groups = self.groups.copy()
        if stage is not None:
            stage.add(newObj)
        # Clone all handlers and handler dictionaries
        newObj._on_click = self._on_click.clone(newObj)
        newObj._on_tick = self._on_tick.clone(newObj)
        newObj._messageHandlers = { msg:cb.clone(newObj) for (msg, cb) in self._messageHandlers.items()}
        newObj._keyHandlers = { msg:cb.clone(newObj) for (msg, cb) in self._keyHandlers.items()}
        
        return newObj
    
    #################################################
    ##                  SENSING
    #################################################
    
    EDGE = 99
    def touching(self, what):
        """
        @param what Can be any of
            * Sprite another sprite - will compare against masks
            * pygame.sprite.Group [ADVANCED]
            * An (x,y) position tuple
            * A pygame.color.Color (but can't be an RGB tuple)
            * Sprite.EDGE to mean the screen edge.
        """
        if isinstance(what, pygame.sprite.Sprite):
            return pygame.sprite.collide_mask(self, what) is not None
        elif isinstance(what, pygame.sprite.Group):
            return len(pygame.sprite.spritecollideany(self, what, pygame.sprite.collide_mask)) > 0
        elif isinstance(what, tuple):  #coordinates
            try:
                return 1 == self._mask.get_at(what[0]-self._rect.left, what[1]-self._rect.top)
            except: # way out of bounds.. TODO optimize no exception?
                return False
        elif isinstance(pygame.color.Color):
            return self.touching_color(what)
        else: #assume EDGE
            return self.touching_edge()
        
    def touching_edge(self):
        #TODO: the mask may not necessarily go to the bounding rectangle
        w, h = get_window().size
        return self._rect.left < 0 \
            or self._rect.top < 0 \
            or self._rect.right >= w \
            or self._rect.bottom >= h
    
    def touching_color(self, color):
        raise NotImplementedError()
        #TODO... how
    
    #TODO: color is touching color??
    
    def distance_to(self, what):
        """
        @param what Can be any of
            * Sprite another sprite - distance center to center
            * An (x,y) position tuple
        """
        def dist(pos1, pos2):
            x1, y1 = pos1
            x2, y2 = pos2
            return math.sqrt((x2-x1)**2 + (y2-y1)**2)
        
        if isinstance(what, pygame.sprite.Sprite):
            return dist(self.position, what.position)
        elif isinstance(what, tuple):  #coordinates
            return dist(self.position, what)
        else:
            raise ValueError("Unknown type given to distance_to()")
        
    def direction_to(self, what):
        """
        Bonus: gives the direction in degrees from this Sprite to the
        given thing.
        @param what Can be any of
            * Sprite another sprite - distance center to center
            * An (x,y) position tuple
        @return a direction in the range [-180,180), where 0 means the
                other object is directly above this one.
        """
        if isinstance(what, Sprite):
            posX = what.x
            posY = what.y
        else:
            posX = what[0]
            posY = what[1]

        dy = posY - self._y # may be negatives
        dx = posX - self._x
        if dx == 0:  # straight up or down
            return 0 if dy >= 0 else 180
        
        rotation = math.atan(dy/dx) * 180 / math.pi
        if dx < 0:
            rotation += 180
        return (rotation + 270) % 360 - 180
        
    async def ask_and_wait(self, whatisyourname):
        """
        @return answer as a string
        """
        self.say(whatisyourname)
        answer = await scratchypy.get_stage()._ask_prompt()
        self.say(None)
        return answer
    
    # key_pressed on window
    # mouse_down, x, y on window
    
    def set_draggable(self, canDrag):
        self._draggable = canDrag
    
    #################################################
    ##                  MISC extras
    #################################################
    def set_debug(self, onoff=True):
        self._debug = onoff
        
    def _render(self, screen):
        if self._visible:
            screen.blit(self._image, self._rect)
            if self._sayThinkImages:
                bubbleRect = self._sayThinkImages[0].get_rect() # assume same size
                # Y no higher than top of screen
                bubbleY = max(0, self._rect.top - bubbleRect.h)
                # Bubble on right side if fits, else left
                if self._rect.right + bubbleRect.w <= screen.get_size()[0]:
                    r = bubbleRect.move(self._rect.right, bubbleY)
                    screen.blit(self._sayThinkImages[0], r) #right
                else:
                    r = bubbleRect.move(self._rect.left - bubbleRect.w, bubbleY)
                    screen.blit(self._sayThinkImages[1], r) #left
            if self._debug:
                pygame.draw.rect(screen, color.BLUE, self._rect, width=1)
                pygame.draw.line(screen, color.GREEN, (self._x-5, self._y), (self._x+5, self._y))
                pygame.draw.line(screen, color.GREEN, (self._x, self._y-5), (self._x, self._y+5))
    blit = _render #XXX
        
    def update(self):
        """
        Called once per frame to do updates.
        """
        self._on_tick()
    
    
class TextSprite(Sprite):
    """
    An extension of a Scratchy Sprite to display text dynamically.
    The text is rendered to a surface and stored as the one and only
    costume.
    This is a simple-to-use sprite on top of pygame.font.
    """
    def __init__(self, text, color=color.BLACK, 
                 font=None, size=50,
                 x=0, y=0, topleft=None, topright=None,
                 justification='left',
                 name=None, maxWidth=600,
                 bgcolor=None, stage=None):
        """
        If x,y is given, the sprite is positioned from center like normal.
        If topleft is given, the left edge remains constant when text changes.
        If topright is given, the right edge remains constant.
        The 'justification' (left, right, center) applies to text that has or
        wraps to multiple lines, and is the alignment within the sprite.
        If a custom font is given, then 'size' is ignored.
        """
        self._maxWidth = maxWidth
        self._color = color
        self._size = size
        self._font = font if font else pygame.font.SysFont("sans serif", self._size)
        self._topleft = topleft
        self._topright = topright
        self._justification = justification
        self._bgcolor = bgcolor
        surface = self._render_text(text)
        Sprite.__init__(self, surface, x=x, y=y, topleft=topleft, name=name, stage=stage)
    
    def _render_text(self, text):
        """
        """
        bounds = pygame.Rect(0,0,self._maxWidth,320)
        return scratchypy.text.render_text(self._font, text, bounds, self._color, self._bgcolor, self._justification)
    
    def set_text(self, text):
        surface = self._render_text(text)
        self._costumes = [ surface ]
        self._applyImage()
        # readjust if we're justified
        if self._topleft:
            self.go_rect(topleft=self._topleft)
        if self._topright:
            self.go_rect(topright=self._topright)
        
        
class AnimatedSprite(Sprite):
    """
    A convenience Sprite that can animate its costumes, assuming
    one frame per costume.  Functions are provided for easy start
    and stop of the animation.  This sprite inherits from Sprite
    and you can still move it and do any other Sprite actions.
    """
    def __init__(self, costumes, fps:int=5, x=0, y=0, name=None, size=None, stage=None):
        """
        @param costumes: List of frame/costume images.  See Sprite doc.
        @param fps: Frames Per Second you want the animation to play at,
               since e.g. a full 30 FPS is usually too fast for simple
               animations.  Will be rounded to the nearest rate compatible
               with the overall framerate.
               
        """
        super().__init__(costumes, x=x, y=y, name=name, size=size, stage=stage)
        FPS = get_window().fps
        if fps <= 0 or fps > FPS:
            raise ValueError("fps must be 1..FPS")
        self._framesPerTick = FPS // fps
        self._frameCounter = 0
        self._playing = True
        
    def start(self):
        self._playing = True
        
    def pause(self):
        self._playing = False
        
    def stop(self, homeFrame = 0):
        self._playing = False
        self.switch_costume_to(homeFrame)
    
    def update(self):
        """
        Called once per frame to do updates.
        """
        super().update()
        if self._playing:
            self._frameCounter += 1
            if self._frameCounter % self._framesPerTick == 0:
                self.next_costume()
        
        
        


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
""" Common named colors """

import pygame
BLACK = pygame.Color(0,0,0)
WHITE = pygame.Color(0xff,0xff,0xff)
GREEN = pygame.Color(0,255,0)
BLUE = pygame.Color(0,0,255)
RED = pygame.Color(255,0,0)
YELLOW = pygame.Color(255,255,0)
ORANGE = pygame.Color(255,128,0)
PURPLE = pygame.Color(255,0,255)


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms.
"""
Contains functions for loading image files.
The file types it supports are the same as what pygame.image supports.
"""

import glob
import pygame.image

def load(fileName, colorToMakeTransparent:pygame.color.Color=None):
    surface = pygame.image.load(fileName)
    if colorToMakeTransparent:
        surface.convert()
        surface.set_colorkey(colorToMakeTransparent)
    else:
        # support per-pixel alpha
        surface.convert_alpha()
    return surface

def loadAll(listOfFiles, transparentColor:pygame.color.Color=None):
    return [ load(f, transparentColor) for f in listOfFiles ]

def loadPattern(globPattern, transparentColor:pygame.color.Color=None):
    """
    Load a bunch of files that match the given C{globPattern}, which may contain
    stars and other special characters as wildcards, e.g. 'pics/rocket*.png'.
    This is useful for loading many similarly-named files for animations.
    The files will be sorted to keep the animation in order (assuming numbers often
    used as the wildcard, but this is not a fancy sort.  If you have more than 10 images,
    always use two digits e.g. a01.png, rather than a1.png, a10.png, a2.png, a3.png, ...
    """
    listOfFiles = glob.glob(globPattern)
    listOfFiles.sort()
    return [ load(f, transparentColor) for f in listOfFiles ]
    


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
"""
Contains the EventCallback class for seamlessly encapsulating many different
types of callback strategies.
"""

import asyncio
import traceback
from scratchypy.util import is_ui_thread

class EventCallback:
    '''
    Encapsulates a callback and the calling of it.
    It handles these cases:
    * callback is None; do nothing
    * callback is async; create a task for it and run it on the event loop
    * callback is a regular function; call it synchronously on the next loop
    * caller has annotated with @to_thread; call it on threadpool
    Exceptions from the user callback are caught and logged.
    TODO: Also supports detection of args and kwargs as optional.??
    
    '''

    def __init__(self, obj, cb, name=''):
        '''
        @param obj The object initiating the callback (e.g. a certain stage or sprite).
               Usually this doesn't change for the lifetime of the callback setting.
        @param cb The callback.  May be a regular or async function, or None for no action.
        @param name Optional name used for debugging.  If omitted, this will try to find
               the __name__ of the callback function.
        '''
        self._obj = obj
        self._task = None
        self._name = name if name else cb.__name__ if cb else '(none)'
        self.set(cb)
        
    def clone(self, newObj):
        """
        Clone this handler to make a copy using the new object.  Used during
        Sprite cloning to attach the same callbacks to the new object.
        Any existing tasks are not cloned.
        """
        return EventCallback(newObj, self._cb, self._name)
        
    def name(self):
        """ @return the name given to this callback, mostly for debugging """
        return self._name
            
    def set(self, cb):
        """
        Set/reset the callback to the given function 'cb'.
        If an existing previous callback is in progress, it is cancelled.
        @param cb A regular or async function.  See class doc for details.
               May also be None.
        """
        self._cb = cb
        
        if self._task:
            print("Resetting callback '%s' while it was running! Canceling" % self._name)
            self._task.cancel()
            self._task = None
            
        if cb is None:
            self._doCall = self._call_noop
        elif asyncio.iscoroutinefunction(self._cb):
            self._doCall = self._call_async
        elif hasattr(self._cb, "_to_thread"):
            self._doCall = self._call_threaded
        else:
            self._doCall = self._call_sync
            
    def _call_noop(self, *args):
        "Do nothing when callback is None"
        pass
            
    def _call_sync(self, *args):
        """
        Call the callback synchronously while handling exceptions.
        Ensures we're on the UI thread as a safety measure.
        """
        if not is_ui_thread():
            raise Exception("Synchronous callback not from UI thread")
    
        # Consistently call within event loop;
        # prevents race condition if `when_started` is async
        # and the tick would happen before initialization.
        # BUT, this makes the callback delayed slightly
        asyncio.get_running_loop().call_soon(
            self._safe_call_sync, *args  ##FIXME: no kwargs
        )
        
    def _call_async(self, *args):
        """
        Call an 'async' callback as an asyncio task.  Returns immediately after the task is scheduled.
        """
        if self._task:
            print("Async callback '%s' still in progress, canceling" % self._name)
            self._task.cancel()
        self._task = asyncio.create_task(self._safe_call_async(*args), name=self._name)
        
    def _call_threaded(self, *args):
        """
        Kick the callback to a separate thread and treat that as the task.
        """
        if self._task:
            print("Thread callback '%s' still in progress, canceling" % self._name)
            self._task.cancel()
        #Python 3.9 or later would use to_thread()
        self._task = asyncio.get_running_loop().run_in_executor(None, self._safe_call_sync, *args) # no kwargs 
            
    def _safe_call_sync(self, *args):
        "Call the callback synchronously while handling exceptions"
        try:
            self._cb(self._obj, *args)
        except Exception as ex:
            print("Callback error: %s: %s" % (self._name, ex))
            traceback.print_exception(ex, ex, ex.__traceback__)
        self._task = None # no-op for sync case; used for @to_thread
            
    async def _safe_call_async(self, *args):
        """
        Wrapper for calling the async callback and doing the housekeeping of the
        tracked task when finished.
        This ensures it runs in the same slice as the callback, as opposed to
        using add_done_callback(), which posts it additionally and could run later.
        """
        try:
            await self._cb(self._obj, *args)
        except asyncio.exceptions.CancelledError:
            pass
        except Exception as ex:
            print("Callback error: %s: %s" % (self._name, ex))
            traceback.print_exception(ex, ex, ex.__traceback__)
        self._task = None  # ready for next one
        
    def __call__(self, *args):
        """
        How to actually call the callback.
        evt = EventCallback(...)
        evt()   # or more explicitly, evt.call()
        """
        self._doCall(*args)
    call = __call__
    
    
    


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
"""
Misc. utility functions.
"""
import asyncio
import os
import threading
import datetime



async def wait(seconds):
    await asyncio.sleep(seconds)
    
def username():
    return os.getlogin()

def days_since_2000():
    now = datetime.datetime.today()
    y2k = datetime.datetime(2000,1,1)
    return (now-y2k).total_seconds() / 86400

#
# THREADING
#
_uiThreadId = None
def set_ui_thread():
    global _uiThreadId
    _uiThreadId = threading.get_ident()

def is_ui_thread():
    return _uiThreadId == threading.get_ident()

#
# DECORATORS
#
def to_thread(uiFunc):
    # See EventCallback for significance of this
    uiFunc._to_thread = True
    return uiFunc

def ui_only(func):
    def inner():
        if not is_ui_thread():
            raise Exception("%s can only be called from the UI thread" % func.__name__)
        else:
            return func
    return inner





# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
"""
Helpers for rendering text.  Rarely needed to be used directly.
"""
import re
import asyncio
import pygame.surface
from scratchypy import color

def render_text(font, text, rect, color=color.BLACK, bgcolor=None, justification='left'):
    maxWidth = rect.w
    
    # word wrap
    lines = []
    for inLine in re.split('\n', text):
        words = re.split('\\s', inLine)
        i = 1
        while i <= len(words):
            w,_ = font.size(' '.join(words[:i]))
            if w > maxWidth:
                # one word too many, unless it's the only word, might truncate
                if i == 0:
                    lines.append(words[0])
                    words = words[1:]
                else:
                    lines.append(' '.join(words[:i-1]))
                    words = words[i-1:]
                i=0
            else:
                i+=1
        # add remaining
        if words:
            lines.append(' '.join(words))
    
    # calculate size of single render surface
    surfaces = [font.render(line, True, color) for line in lines]
    # If only one and we can return it as-is.
    if len(surfaces) == 1 and not bgcolor:
        return surfaces[0]  #TODO: may not honor maxHeight
    # else blit them all to a single surface
    width = min(maxWidth, max([s.get_width() for s in surfaces]))
    totalHeight = min(rect.h, font.get_linesize() * len(surfaces))
    #TODO: nicer cutoff of lines^^^
    bigsurf = pygame.surface.Surface((width, totalHeight), pygame.SRCALPHA, 32)
    if bgcolor:
        bigsurf.fill(bgcolor)
    
    # render the lines
    y = 0
    for surf in surfaces:
        # Align based on justification.  For long words, may truncate
        x = width - surf.get_width() if justification=='right' else \
            (width - surf.get_width())/2 if justification=='center' else \
            0
        bigsurf.blit(surf, (x,y))
        y += font.get_linesize()
    return bigsurf


class AskDialog(pygame.sprite.Sprite):
    """
    Shows a text box and handles key events, while rendering them to
    the UI as typed.  Only one line is allowed and RETURN submits the
    line.  Callers should use this like a sprite, then await on the
    done() future to get the string that was typed.
    """
    def __init__(self, rect):
        super().__init__()
        self.name = "askModalDialog"  # conform with ScratchyPy sprite
        self._doneFuture = asyncio.Future()
        self._text = ""
        self._draggable = False
        scr = rect
        self.rect = pygame.Rect(10, scr.h-40, scr.w-20, 30)
        self._paddedText = self.rect.inflate(-15,-15)
        self._font = pygame.font.SysFont("sans serif", 24)
    
    def done(self):
        return self._doneFuture
    
    def _on_key_down(self, event):
        if event.key == K_RETURN:
            self._doneFuture.set_result(self._text)
        elif event.key == K_ESCAPE:
            self._doneFuture.cancel()
        else:
            self._text += event.unicode
        # else ignore
        while self._font.size(self._text)[0] > self._paddedText.w:
            self._text = self._text[:-1]
    
    def _render(self, screen):
        pygame.draw.rect(screen, color.WHITE, self.rect)
        pygame.draw.rect(screen, color.BLACK, self.rect, width=2, border_radius=2)
        surf = self._font.render(self._text, True, color.BLACK)
        screen.blit(surf, self._paddedText)






# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 

import sys
sys.path.append("..")
from scratchypy import *
from scratchypy import color

class TitleScreen(Stage):
    def on_init(self):
        # Convenience: specifying the stage automatically adds it.
        self._sprite1 = Sprite("assets/axolotl1.png", 
                         x=400,y=100,
                         size=50, 
                         name='ax1', stage=self)
        
        # Or the stage can be added explicitly.
        self._title = TextSprite("The Most Amazing Game Ever", x=400, y=300, maxWidth=800)
        self.add(self._title)
        
        # how to add an event handler for class methods
        self._startButton = TextSprite("Start", x=400, y=400, bgcolor=color.GREEN, stage=self)
        self._startButton.when_clicked(self.onStart)
        
    def onStart(self, sender, pos):
        set_stage(GameScreen())

class GameScreen(Stage):
    def on_init(self):
        self._sprite1 = Sprite("assets/axolotl1.png", 
                         x=300,y=300,
                         size=50, 
                         name='ax1',
                         stage=self)
        self.forever(self.tick)
        self.gravity = 1
        
    def tick(self, stage):
        # 'stage' arg ^^ is redundant
        self._sprite1.change_y_by(self.gravity)
        self.gravity *= 1.1
        if self._sprite1.touching_edge():
            set_stage(EndScreen())

class EndScreen(Stage):
    def on_init(self):
        self._title = TextSprite("End stage:\nOoops you died", x=400, y=300, maxWidth=800, justification='center', stage=self)

if __name__ == '__main__':
    start(stage=TitleScreen())
    
    
    


# Copyright 2023 Mark Malek
# See LICENSE file for full license terms. 

import sys
sys.path.append("..")
from scratchypy import *
from scratchypy import color

class MainScreen(Stage):
    def on_init(self):
        # Convenience: specifying the stage automatically adds it.
        self._sender = Sprite("assets/axolotl1.png", 
                         x=100,y=300,
                         size=50, 
                         name='sender', stage=self)
        # Set handler for "hello" message
        self._sender.when_i_receive("hello", self.helloMessageHandler)
        self._sender.when_i_receive("broadcast", self.broadcastHandler)
        self._receiver1 = self._sender.clone("receiver1", self)
        self._receiver2 = self._sender.clone("receiver2", self)
        self._receiver3 = self._sender.clone("receiver3", self)
        
        self.when_started(self._async_start)
        
    async def _async_start(self, obj):
        t1 = self._receiver1.glide_to(500, 100, 2)
        t2 = self._receiver2.glide_to(500, 300, 2)
        t3 = self._receiver3.glide_to(500, 500, 2)
        await t1
        await t2
        await t3
        
        self._sender.when_clicked(self.onSenderClicked)
        self._sender.say("Click me to send a message to the first clone.")
        
    def onSenderClicked(self, sprite, pos):
        sprite.say()
        self._receiver1.message("hello", {'senderName':"Axel Ottle"})
        
    async def helloMessageHandler(self, sprite, argDictionary):
        await sprite.say_and_wait("Why hello %s" % argDictionary.get("senderName", "??"), 2)
        sprite.when_clicked(self.onReceiver1Clicked)
        sprite.say("Click me to broadcast a message to everyone else.")
        
    def onReceiver1Clicked(self, sprite, pos):
        # broadcasts happen from the stage only
        self.broadcast("broadcast", excludeOriginator=sprite)
        sprite.say()

    async def broadcastHandler(self, sprite, argDictionary):
        await sprite.say_and_wait("Roger that.", 3)

if __name__ == '__main__':
    start(stage=MainScreen())
    
    
    


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 

import sys, random
sys.path.append("..")
from scratchypy import *
from pygame.locals import *

sp1 = None
txt = None

async def init(stage):
    global sp1, sp2, txt
    print("init")
    
    stage.add_backdrops(main="assets/background1.jpg")
    stage.forever(tick)
    stage.when_clicked(stageClick)
    stage.when_key_pressed('a', stageKey)
    stage.when_key_pressed('x', changeBg)
    stage.when_drawing(drawExtra)
    
    
    frames = image.loadPattern("assets/axolotl*.png")
    # Make the tail wag both ways
    frames.extend(reversed(frames[:-1]))
    sp1 = AnimatedSprite(frames, 5, 
                         x=90,y=90,
                         size=50, 
                         name='ax1')
    stage.add(sp1) 
    sp1.when_clicked(onClick)
    sp1.when_i_receive("hello", onHello)
    sp1.when_key_pressed('b', spriteKey)
    
    txt = scratchypy.sprite.TextSprite("Hello there to all the cool axolotls with sunglasses", x=320, y=220)
    stage.add(txt)
    
    sp2 = AnimatedSprite(frames, 1, 
                         x=30,y=550,
                         size=20, 
                         name='ax2')
    stage.add(sp2)
    #Story mode
    sp2.glide_to(30, 450, 1) #background
    await sp2.say_and_wait("swimmin along", 1)
    await sp2.glide_to_position_and_wait((700, 550), 5)
    print("init done")

def tick(stage):
    global sp1
    #print("tick")
    #if random.randint(1,10) > 6:
    #    sp1.turn(random.randint(-40,40))
    #sp1.setRotationStyle(scratchypy.LEFT_RIGHT)
    w = get_window()
    sp1.point_towards((w.mouse_x, w.mouse_y))
    #sp1.change_size_by(random.randint(-5, 5))
    sp1.move(3)
    #sp1.if_on_edge_bounce()
    
    txt.move(1)
    txt.turn(2)
    #stage.broadcast("hello", who="worldx")
    #print(pygame.key.name(pygame.K_UP))
    if w.key_pressed("up"):
        sp1.change_y_by(-30)
    if w.key_pressed("down"):
        sp1.change_y_by(30)
    if w.key_pressed("c"):
        print(sp1.y_position)
        sp2 = sp1.clone("sp2", stage)
        sp2.glide_to(0,0,5)
        
async def stageKey(stage):
    #sp1.change_x_by(30)
    print("stageKey")
    answer = await sp1.ask_and_wait("what you want?")
    print(answer)

async def spriteKey(src):
    #sp1.change_x_by(-30)
    print("sleep")
    #await asyncio.sleep(3)
    await sp1.think_and_wait("ouch that hurt a lot", 2)
    print("wake")
    #raise Exception()

@to_thread
def onClick(relpos):
    print("HIT ", relpos)
    
def stageClick(pos):
    print("Miss")
    print("FPS: %f" % get_window().actual_fps)
    print(sp1.direction_to((200,200)))
    
def changeBg(stage):
    stage.next_backdrop()
    
def drawExtra(stage, surface):
    pygame.draw.rect(surface, color.BLUE, pygame.Rect(200,200,2,2), width=1)
    

def onHello(**kwargs):
    print(kwargs["who"])

if __name__ == '__main__':
    start(init)
    
    
    


# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
"""
Contains the Stage class used as a canvas to draw Sprites upon.
"""

import inspect
import random
from typing import Union
import pygame
from scratchypy.eventcallback import EventCallback
import scratchypy.window 
from scratchypy.text import AskDialog

#TODO prefer composition
#FIXME: "The Group does not keep sprites in any order"; we want ordering for layers
class Stage(pygame.sprite.Group):
    '''
    A Stage is the main element that contains all the Sprites and dispatches
    events to the Sprites.  It also contains a backdrop image.
    '''

    def __init__(self):
        '''
        Constructor
        '''
        pygame.sprite.Group.__init__(self)
        self._on_start = EventCallback(self, None, name="Stage.when_started")
        self._on_tick = EventCallback(self, None, name="Stage.each_tick")
        # list of (name, surface) tuples
        self._backdrops = []
        self._backdropId = -1 #TODO: gotta be one default
        self._name_lookup = {}
        self._on_click = EventCallback(self, None, name="Stage.when_clicked")
        self._allClickEvents = False
        # Dict of key->handler
        self._keyHandlers = {}
        self._dialog = None
        self._draw_raw = EventCallback(self, None, name="Stage.when_drawing")
        # call subclass init
        self.on_init()
        
    def on_init(self):
        """
        If you inherit from Stage to group your scenes and sprites, override
        this to create and add your sprites.
        This is slightly easier than having to make your own __init__() and
        needing to call super().__init__(), although that is still valid.
        TODO: init vs start is confusing.  Make it work for both 'when_started'
        and subclassing, and for both sync and async...
        """
        pass
        
    def _start(self):
        self._on_start()
        
    def _update(self, screen):
        if self._backdropId >= 0:
            screen.blit(self._backdrops[self._backdropId][1], (0,0))
        self._on_tick()
        #pygame.sprite.Group.update(self)  XXX
        for sprite in self.sprites():
            sprite.update()
            sprite._render(screen)
        self._draw_raw(screen)
        if self._dialog:
            self._dialog._render(screen)
            
            
    def _on_mouse_down(self, event):
        #TODO: distinguish click vs. drag
        pass
            
    def _on_mouse_motion(self, event):
        for sp in self.sprites():
            if sp._draggable:
                sp._on_mouse_motion(event)
                
    def _on_mouse_up(self, event):
        "A click happens on mouse up"
        # Event will contain fields pos, button, touch
        if event.button != 1:
            return # Don't handle right clicks now
        # Find the sprite(s) that has the position
        handled = False
        for sp in self.sprites():
            spPos = (int(event.pos[0] - sp._rect.left), int(event.pos[1] - sp._rect.top))
            #TODO: check for mask hit or is rect enough?
            if sp._rect.collidepoint(event.pos) and sp._mask.get_at(spPos) == 1:
                handled=True
                # Hit - call the sprite's handler
                sp._on_click(spPos) # Todo: what params to pass?
        # send event to the stage if registered
        if self._on_click and (self._allClickEvents or not handled):
            self._on_click(event.pos)
            
    def _on_key_down(self, event):
        if self._dialog:
            self._dialog._on_key_down(event)
            return # make it modal
        
        handler = self._keyHandlers.get(event.key)
        if handler:
            handler(self)
        # Pass to sprites too
        for sp in self.sprites():
            sp._on_key_down(event)
            
    def add_backdrop(self, image:Union[str,pygame.Surface], name:str=None):
        """
        Add a background image to the stage.  If this is the first
        one added, it will be automatically switched to.
        @param image Either a filename or an already converted Surface
        @param name (optional) A name to use for this backdrop.  If omitted,
               then a name is generated from its 0-based index.
        """
        if isinstance(image, str):
            im = pygame.image.load(image)
            im.convert()
            #TODO: scale image to window
        elif isinstance(image, pygame.Surface):
            im = image
        else:
            raise TypeError("I don't know what this backdrop is")
        # Resize to screen.  TODO: may warp
        im = pygame.transform.smoothscale(im, scratchypy.window.get_window().size)
        
        name = name if name else "backdrop" + str(len(self._backdrops))
        self._backdrops.append((name, im))
        if self._backdropId < 0:
            self.switch_backdrop_to(0)
        
    def add_backdrops(self, *backdrops, **kwBackdrops):
        """
        Adds several backdrops at once.
        @param backdrops Positional arguments adding an arbitrary number of
               backdrops in order, of things compatible with add_backdrop()
        @param kwBackdrops Keyword arguments adding backdrops with the name as 
               the key, and a value compatible with add_backdrop().
        """
        for b in backdrops:
            self.add_backdrop(b)
        for k,v in kwBackdrops.items():
            self.add_backdrop(v, name=k)
    
    def add(self, *sprites):
        # overrides Group impl to also track names
        # TODO: handle collisions?
        for sp in sprites:
            self._name_lookup[sp.name] = sp
        pygame.sprite.Group.add(self, *sprites)
        
    #################################################
    ##                  LOOKS
    #################################################
    @property
    def backdrop_name(self) -> str:
        if self._backdropId < 0:
            return "(default)"
        
    
    @property
    def backdrop_number(self) -> int:
        """
        Note: starts at 0.  May be -1 if there are no backdrops at all.
        """
        return self._backdropId
    
    def switch_backdrop_to(self, nameOrIndex:Union[str,int]):
        if isinstance(nameOrIndex, str):
            for idx, bdTup in enumerate(self._backdrops):
                if bdTup[0] == nameOrIndex:
                    self._backdropId = idx
            else:
                raise KeyError("No backdrop named " + nameOrIndex)
        elif isinstance(nameOrIndex, int):
            if nameOrIndex < 0 or nameOrIndex >= len(self._backdrops):
                raise IndexError("No backdrop at (0-based) index %d" % nameOrIndex)
            self._backdropId = nameOrIndex
        else:
            raise TypeError("Unknown backdrop id")
        
    async def switch_backdrop_and_wait(self):
        # TODO: should switch, call any callbacks for when switched, and return
        # that callback's future.
        raise NotImplementedError()
        
    def next_backdrop(self):
        if self._backdrops:
            self._backdropId = (self._backdropId + 1) % len(self._backdrops)
            
    def previous_backdrop(self):
        if self._backdrops:
            if self._backdropId < 0:
                self._backdropId = 0 # so it will roll to last image below
            self._backdropId = (self._backdropId - 1) % len(self._backdrops)
            
    def random_backdrop(self):
        if self._backdrops:
            self._backdropId = random.randint(0, len(self._backdrops))
    
    #################################################
    ##                  EVENTS
    #################################################
            
    def when_started(self, functionToCall):
        self._on_start.set(functionToCall)
        
    def when_key_pressed(self, key, functionToCall):
        #TODO cancel previous
        if not inspect.isfunction(functionToCall):
            raise TypeError("callback is not a function")
        
        if isinstance(key, str):
            intkey = pygame.key.key_code(key) #throw if not found
        elif isinstance(key, int):
            intkey = key
        else:
            raise TypeError("unknown key")
        self._keyHandlers[intkey] = EventCallback(self, functionToCall)
        
    def forever(self, functionToCall):
        self._on_tick.set(functionToCall)
    each_tick = forever
        
    def when_clicked(self, handler, allClicks=False):
        """
        When a any part of the stage is clicked, call the given handler
        If allClicks=True, this will be called even if there are sprites at this point too.
        """
        if not inspect.isfunction(handler):
            raise TypeError("callback is not a function")
        
        self._on_click = EventCallback(self, handler)
        self._allClickEvents = allClicks
        
    def broadcast(self, messageName, argDictionary={}, excludeOriginator=None):
        for sp in [sp for sp in self.sprites() if sp is not excludeOriginator]:
            sp.message(messageName, argDictionary)
            
            
    #################################################
    ##                  SENSING
    #################################################
    # TODO keypress - should go to sprites too??
    
    async def _ask_prompt(self):
        """
        Draws the text box on the string and asynchronously waits until 
        the user has entered the value.
        @return answer as a string
        """
        self._dialog = AskDialog(scratchypy.window.get_window().rect)
        try:
            answer = await self._dialog.done()
        finally:
            self._dialog = None
        return answer
    
    #################################################
    ##                  VARIABLES
    #################################################
    # TODO: show variable?  or leave it to a debugger?
            
    #################################################
    ##                  BONUS
    #################################################
    def when_drawing(self, callback):
        """
        Register an event handler to be called when the screen
        and all sprites are done drawing.  This gives a chance
        to manually draw more stuff with raw pygame commands
        on the raw pygame Surface.
        The callback should look like this:
        def drawExtraStuff(stage, surface):
            pygame.draw.rect(...)  # e.g.
        """
        self._draw_raw.set(callback)